/* Copyright (c) 2000 John Stracke <fanorona@thibault.org>
 #
 * This file is part of Fanorona.
 #
 * Fanorona is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any later
 * version. 
 #
 * Fanorona is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details. 
 #
 * You should have received a copy of the GNU General Public License along
 * with program; see the file COPYING. If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <errno.h>
#include <gnome.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"

#include "staticWindows.h"
#include "prefs.h"

static int curDepth=3;
static int curDelay=500;
static const char* colorSetName=0;
static int colorSetIndex=-1;
static int colorSetIndexTentative=-1;

static const int delays[3]={2000,1000,500};

static struct {
  int curDepth;
  int colorSetIndex;
  int curDelay;
} savedSettings;

typedef struct {
  const char* empty;
  const char* player1;
  const char* player2;
  const char* borders;
  const char* pressed;
  const char* possible;
} ColorSet;

static struct {
  const char* name;
  ColorSet colors;
} colorSets[]={
  {"Red/Green",{"white","red","green","black","yellow","yellow"}},
  {"Yellow/Blue",{"white","yellow","blue","black","red","red"}},
  {"Theme",{"","","","","",""}}
};

static const ColorSet* getColorSet(const char* name,
                                   int* index)
{
  int i,N;
  for (i=0, N=sizeof(colorSets)/sizeof(colorSets[0]); i<N; i++)
    {
      if (!strcmp(name,colorSets[i].name))
        {
          if (index)
            (*index)=i;
          return &(colorSets[i].colors);
        }
    }

  return 0;
}

static void applyColorSet(const ColorSet* colors,
                          gboolean invalidate)
{
  gtk_fanorona_set_colors(getFanorona(),
                          colors->empty,
                          colors->player1,
                          colors->player2,
                          colors->borders,
                          colors->pressed,
                          colors->possible,
                          invalidate);
}

void handlePrefs(void)
{
  ColorSet colors={"white","red","green","black","yellow","yellow"};

  curDepth=getPrefInt("depth",3);

  curDelay=getPrefInt("delay",3);
  gtk_fanorona_set_delay(getFanorona(),curDelay);
  
  colorSetName=getPrefWithDef("colorset","Red/Green");
  if (colorSetName)
    {
      const ColorSet* colorSet=getColorSet(colorSetName,&colorSetIndex);
      if (colorSet)
        colors=(*colorSet);
    }

  colors.empty=getPrefWithDef("colors.empty",colors.empty);
  colors.player1=getPrefWithDef("colors.player1",colors.player1);
  colors.player2=getPrefWithDef("colors.player2",colors.player2);
  colors.borders=getPrefWithDef("colors.borders",colors.borders);
  colors.pressed=getPrefWithDef("colors.pressed",colors.pressed);
  colors.possible=getPrefWithDef("colors.possible",colors.possible);

  applyColorSet(&colors,0);
}

static void set_status(const char* status)
{
  GnomeApp* app=GNOME_APP(getMainWindow());
  GnomeAppBar* bar=GNOME_APPBAR(app->statusbar);

  gnome_appbar_set_status(bar,status);
}

static void clear_status(void)
{
  set_status("");
}

FanoronaPlayer* getFanoronaPlayer(void)
{
  static FanoronaPlayer* player=0;
  if (!player)
    player=fanoronaPlayerNew(0,curDepth);
  return player;
}

static GtkWidget* getPrefsDepthSpinbutton(void)
{
  GtkWidget* prefsDialog=getPrefsDialog();
  if (!prefsDialog)
    return 0;

  return lookup_widget(prefsDialog,"prefsDepthSpinbutton");
}

static GtkOptionMenu* getPrefsColorSetOptionMenu(void)
{
  GtkWidget* prefsDialog=getPrefsDialog();
  if (!prefsDialog)
    return 0;

  return GTK_OPTION_MENU(lookup_widget(prefsDialog,"colorSetOptionMenu"));
}

void
on_new_game_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gtk_fanorona_new_game(getFanorona());
}


void
on_exit_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gtk_exit(0);
}


void
on_undo_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gtk_fanorona_undo(getFanorona());
}

void
on_redo_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gtk_fanorona_redo(getFanorona());
}

static void on_colorSetButton_activate(GtkButton* button,
                                      gpointer _i)
{
  colorSetIndexTentative=(*((int*)_i));
}

void
on_preferences_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget* prefsDialog=getPrefsDialog();
  GtkWidget* prefsDepthSpinbutton=0;

  if (!prefsDialog)
    return;

  prefsDepthSpinbutton=getPrefsDepthSpinbutton();
  if (prefsDepthSpinbutton)
    gtk_spin_button_set_value(GTK_SPIN_BUTTON(prefsDepthSpinbutton),
                              curDepth);

  if (colorSetIndex>=0)
    {
      GtkOptionMenu* prefsColorSetOptionMenu=getPrefsColorSetOptionMenu();
      static gboolean inited=0;
      if (!inited)
        {
          GtkMenu* menu
            =GTK_MENU(gtk_option_menu_get_menu(prefsColorSetOptionMenu));
          GList* children=menu->menu_shell.children;
          int i;

          for (i=0; children; i++, children=children->next)
            {
              GtkWidget* w=(GtkWidget*)(children->data);
              GtkMenuItem* menuItem=GTK_MENU_ITEM(w);
              int* p=(int*)malloc(sizeof(int));
              (*p)=i;
              gtk_signal_connect(GTK_OBJECT(menuItem),
                                 "activate",
                                 GTK_SIGNAL_FUNC(on_colorSetButton_activate),
                                 p);
            }

          inited=1;
        }

      gtk_option_menu_set_history(prefsColorSetOptionMenu,colorSetIndex);
      colorSetIndexTentative=-1;
    }

  {
    GtkToggleButton* slow
      =GTK_TOGGLE_BUTTON(lookup_widget(prefsDialog,"speedSlowRadioButton"));
    GtkToggleButton* medium
      =GTK_TOGGLE_BUTTON(lookup_widget(prefsDialog,"speedMediumRadioButton"));
    GtkToggleButton* fast
      =GTK_TOGGLE_BUTTON(lookup_widget(prefsDialog,"speedFastRadioButton"));

    GtkToggleButton* cur;

    if (curDelay<=delays[2])
      cur=fast;
    else
      {
        if (curDelay>=delays[0])
          cur=slow;
        else
          cur=medium;
      }

    gtk_toggle_button_set_active(cur,1);
  }

  savedSettings.curDepth=curDepth;
  savedSettings.curDelay=curDelay;
  savedSettings.colorSetIndex=colorSetIndex;

  gtk_widget_show (prefsDialog);
}


void
on_about_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget* about=getAboutBox();
  if (!about)
    return;

  gtk_widget_show (about);
}


gint
on_aboutBox_close                      (GnomeDialog     *gnomedialog,
                                        gpointer         user_data)
{
  printf("on_aboutBox_close()\n");
  gtk_widget_hide(GTK_WIDGET(gnomedialog));
  return 1;
}

void game_over(GtkFanorona* fanorona,
               gboolean player1Won,
               gpointer data)
{
  GtkWidget* dialog=0;

  printf("%s win.\n",(player1Won ? "You" : "I"));

  if (player1Won)
    dialog=create_youWinBox();
  else
    dialog=create_iWinBox();

  gtk_widget_show(dialog);
}

void prompt_for_direction(GtkFanorona* fanorona,
                          gpointer data)
{
  set_status("By approach or by withdrawal?");
}

void remove_prompt(GtkFanorona* fanorona,
                   gpointer data)
{
  clear_status();
}

void report_whose_move(GtkFanorona* fanorona,
                       gboolean player1sMove)
{
  if (player1sMove)
    set_status("Your move.");
  else
    set_status("Computer's move.");
}

static void prefsSet(GnomeDialog* prefsDialog)
{
  GtkWidget* prefsDepthSpinbutton=0;

  printf("prefsSet()\n");

  prefsDepthSpinbutton=getPrefsDepthSpinbutton();
  if (prefsDepthSpinbutton)
    {
      GtkSpinButton* spinner=GTK_SPIN_BUTTON(prefsDepthSpinbutton);
      curDepth
        =gtk_spin_button_get_value_as_int(spinner);
      fanoronaPlayerSetDepth(getFanoronaPlayer(),curDepth);
      setPrefInt("depth",curDepth);
    }

  if (colorSetIndexTentative>=0)
    {
      colorSetIndex=colorSetIndexTentative;
      setPref("colorset",colorSets[colorSetIndex].name);
      applyColorSet(&(colorSets[colorSetIndex].colors),1);
    }

  gtk_fanorona_set_delay(getFanorona(),curDelay);
  setPrefInt("delay",curDelay);

  savePrefs();
}

static void prefsRevert(GnomeDialog* prefsDialog)
{
  curDepth=savedSettings.curDepth;
  curDelay=savedSettings.curDelay;
  colorSetIndex=savedSettings.colorSetIndex;

  fanoronaPlayerSetDepth(getFanoronaPlayer(),curDepth);
  setPrefInt("depth",curDepth);

  gtk_fanorona_set_delay(getFanorona(),curDelay);
  setPrefInt("delay",curDelay);

  setPref("colorset",colorSets[colorSetIndex].name);
  applyColorSet(&(colorSets[colorSetIndex].colors),1);

  savePrefs();
}

static void prefsClosed(GnomeDialog* prefsDialog)
{
  printf("prefsClosed()\n");
  gtk_widget_hide(GTK_WIDGET(prefsDialog));
}

void
on_prefsDialog_clicked                 (GnomeDialog     *gnomedialog,
                                        gint             arg1,
                                        gpointer         user_data)
{
  printf("on_prefsDialog_clicked(%d)\n",arg1);

  switch (arg1)
    {
    case 0:
      prefsSet(gnomedialog);
      prefsClosed(gnomedialog);
      break;

    case 1:
      prefsSet(gnomedialog);
      break;

    case 2:
      prefsRevert(gnomedialog);
      prefsClosed(gnomedialog);
      break;

    default:
      printf("on_prefsDialog_clicked(): unknown arg1 value %d\n",arg1);
      break;
    }
}


gint
on_prefsDialog_close                   (GnomeDialog     *gnomedialog,
                                        gpointer         user_data)
{
  printf("on_prefsDialog_close()\n");
  prefsClosed(gnomedialog);
  return 1;
}


void
on_how_to_play_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gnome_help_goto(0,"ghelp:" FANORONA_HELPDIR "/docs.html");
}


gboolean
on_mainWindow_delete_event             (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  gtk_exit(0);
  return FALSE;
}

void
on_speedSlowRadioButton_toggled        (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  if (gtk_toggle_button_get_active(togglebutton))
    curDelay=delays[0];
}


void
on_speedMediumRadioButton_toggled      (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  if (gtk_toggle_button_get_active(togglebutton))
    curDelay=delays[1];
}


void
on_speedFastRadioButton_toggled        (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  if (gtk_toggle_button_get_active(togglebutton))
    curDelay=delays[2];
}

typedef enum {
  fileSelectionModeInvalid=0,
  fileSelectionModeOpen,
  fileSelectionModeSave
} FileSelectionMode;

static FileSelectionMode fileSelectionMode=fileSelectionModeInvalid;
static const char* lastFilename=0;

static void openFileSelection(FileSelectionMode mode)
{
  GtkFileSelection* fileSelection=getFileSelection();

  if (mode==fileSelectionModeOpen)
    gtk_file_selection_set_filename(fileSelection,"");
  fileSelectionMode=mode;

  gtk_widget_show(GTK_WIDGET(fileSelection));
}

void openGame(const char* file)
{
  printf("openGame(%s)\n",file);
  if (gtk_fanorona_restore(getFanorona(),file)<0)
    {
      int tmp=errno;
      fprintf(stderr,"openGame(): can't restore game: %s\n",
              strerror(tmp));
      errno=tmp;
      return;
    }
  lastFilename=file;
}

static void saveGame(const char* file)
{
  printf("saveGame(%s)\n",file);
  if (gtk_fanorona_save(getFanorona(),file)<0)
    {
      int tmp=errno;
      fprintf(stderr,"saveGame(): can't save game: %s\n",
              strerror(tmp));
      errno=tmp;
      return;
    }
  lastFilename=file;
}

void
on_openGame_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  openFileSelection(fileSelectionModeOpen);
}


void
on_save_activate                       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  if (lastFilename)
    saveGame(lastFilename);
  else
    on_save_as_activate(menuitem,user_data);
}


void
on_save_as_activate                    (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  openFileSelection(fileSelectionModeSave);
}


void
on_fileselection_ok_clicked            (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkFileSelection* fileSelection=getFileSelection();

  const char* filename=gtk_file_selection_get_filename(fileSelection);
  if (!(filename && (*filename)))
    return;

  switch (fileSelectionMode)
    {
    case fileSelectionModeOpen:
      openGame(filename);
      break;

    case fileSelectionModeSave:
      saveGame(filename);
      break;

    case fileSelectionModeInvalid:
      fprintf(stderr,"on_fileselection_ok_clicked(): mode is Invalid\n");
      break;

    default:
      fprintf(stderr,"on_fileselection_ok_clicked(): unknown mode\n");
      break;
    }

  gtk_widget_hide(GTK_WIDGET(fileSelection));
}


void
on_fileselection_cancel_clicked        (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_hide(GTK_WIDGET(getFileSelection()));
}

static int resizing=0;

static void resize_board(int width,
                         int height)
{
  GtkWidget* fanoronaWidget=GTK_WIDGET(getFanorona());

  fprintf(stderr,"resize_board(%d,%d)\n",width,height);

  if (!fanoronaWidget)
    {
      fprintf(stderr,"resize_board(): fanoronaWidget is NULL\n");
      return;
    }

  resizing++;
  gtk_widget_set_usize(fanoronaWidget,width,height);
  resizing--;
}

void resize_board_from_prefs(void)
{
  resize_board(getPrefInt("board_width",0),
               getPrefInt("board_height",0));
}

void
on_fanorona_size_allocate              (GtkWidget       *widget,
                                        GtkAllocation   *allocation,
                                        gpointer         user_data)
{
  static int zeroing=0;

  if (resizing>0)
    {
      fprintf(stderr,"on_fanorona_size_allocate(): resizing<=0: %d\n",
              resizing);
      return;
    }

  if (zeroing>0)
    {
      zeroing--;
      return;
    }

  printf("Widget is now %d by %d\n",
         allocation->width, allocation->height);

  setPrefInt("board_width",allocation->width);
  setPrefInt("board_height",allocation->height);
  zeroing++;
  gtk_widget_set_usize(widget,0,0);
  savePrefs();
}


GtkWidget*
new_fanorona (gchar *widget_name, gchar *string1, gchar *string2,
                gint int1, gint int2)
{
  return gtk_fanorona_new (0);
}
