/* Copyright (c) 2000 John Stracke <fanorona@thibault.org>
 #
 * This file is part of Fanorona.
 #
 * Fanorona is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any later
 * version. 
 #
 * Fanorona is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details. 
 #
 * You should have received a copy of the GNU General Public License along
 * with program; see the file COPYING. If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 */

#include <errno.h>
#include <fanorona/board.h>

FanoronaBoardCellState
fanoronaBoardCellOppositeState(FanoronaBoardCellState state)
{
  switch (state)
    {
    case fanoronaBoardCellStatePlayer1:
      return fanoronaBoardCellStatePlayer2;

    case fanoronaBoardCellStatePlayer2:
      return fanoronaBoardCellStatePlayer1;

    case fanoronaBoardCellStateInvalid:
    case fanoronaBoardCellStateEmpty:
    default:
      return fanoronaBoardCellStateInvalid;
    }
}

int fanoronaBoardInit(FanoronaBoard* board)
{
  int col,row;

  if (!board)
    {
      fprintf(stderr,"fanoronaBoardInit(): board is NULL\n");
      errno=EFAULT;
      return -1;
    }

  for (col=0; col<FANORONA_NUM_COLS; col++)
    for (row=0; row<FANORONA_NUM_ROWS; row++)
      board->cells[row][col]=fanoronaBoardCellStateEmpty;

  board->numForPlayer1=0;
  board->numForPlayer2=0;

  board->watchers=0;
  board->blockWatchers=0;
  
  return 0;
}

int fanoronaBoardSetup(FanoronaBoard* board)
{
  int col;
  int humanMustWin=(getenv("HUMAN_MUST_WIN")!=0);

  if (!board)
    {
      fprintf(stderr,"fanoronaBoardSetup(): board is NULL\n");
      errno=EFAULT;
      return -1;
    }

  board->blockWatchers++;

  for (col=0; col<FANORONA_NUM_COLS; col++)
    {
      fanoronaBoardSetCell(board,col,0,fanoronaBoardCellStatePlayer1);
      fanoronaBoardSetCell(board,col,1,fanoronaBoardCellStatePlayer1);
      if ((col!=4) && (!humanMustWin))
        fanoronaBoardSetCell(board,col,2,
                             (
                              ((col+(col>4))%2)
                              ? fanoronaBoardCellStatePlayer1
                              : fanoronaBoardCellStatePlayer2
                              )
                             );
      if ((!humanMustWin) || (col==0)) {
        fanoronaBoardSetCell(board,col,3,fanoronaBoardCellStatePlayer2);
        fanoronaBoardSetCell(board,col,4,fanoronaBoardCellStatePlayer2);
      }
    }

  fanoronaBoardSetCell(board,4,2,fanoronaBoardCellStateEmpty);

  board->blockWatchers--;
}

int fanoronaBoardCountSides(const FanoronaBoard* board,
                            int* numForPlayer1,
                            int* numForPlayer2)
{
  if (!board)
    {
      fprintf(stderr,"fanoronaBoardCountSides(): board is NULL\n");
      errno=EFAULT;
      return -1;
    }

  if (numForPlayer1)
    (*numForPlayer1)=board->numForPlayer1;
  if (numForPlayer2)
    (*numForPlayer2)=board->numForPlayer2;

  return 0;
}

int fanoronaBoardGameOver(const FanoronaBoard* board)
{
  int numForPlayer1,numForPlayer2;
  if (fanoronaBoardCountSides(board,&numForPlayer1,&numForPlayer2)<0)
    {
      int tmp=errno;
      fprintf(stderr,"fanoronaBoardGameOver(): can't count sides: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  return ((!numForPlayer1) || (!numForPlayer2));
}

static void triggerWatchers(FanoronaBoard* board,
                            int col,
                            int row,
                            FanoronaBoardCellState state)
{
  FanoronaBoardWatcher cur;

  if (!board)
    return;

  if (board->blockWatchers)
    return;

  for (cur=board->watchers; cur; cur=cur->next)
    {
      if (!(cur->callback))
        continue;

      (cur->callback)(board,col,row,state,cur->context);
    }
}

int fanoronaBoardSetCell(FanoronaBoard* board,
                         int col,
                         int row,
                         FanoronaBoardCellState state)
{
  if (!board)
    {
      fprintf(stderr,"fanoronaBoardSetCell(): board is NULL\n");
      errno=EFAULT;
      return -1;
    }

  if (row<0)
    {
      fprintf(stderr,"fanoronaBoardSetCell(): row<0\n");
      errno=EINVAL;
      return -1;
    }

  if (row>=FANORONA_NUM_ROWS)
    {
      fprintf(stderr,"fanoronaBoardSetCell(): row>%d\n",
              FANORONA_NUM_ROWS);
      errno=EINVAL;
      return -1;
    }

  if (col<0)
    {
      fprintf(stderr,"fanoronaBoardSetCell(): col<0\n");
      errno=EINVAL;
      return -1;
    }

  if (col>=FANORONA_NUM_COLS)
    {
      fprintf(stderr,"fanoronaBoardSetCell(): col>%d\n",
              FANORONA_NUM_COLS);
      errno=EINVAL;
      return -1;
    }

  switch (state)
    {
    case fanoronaBoardCellStateEmpty:
      break;

    case fanoronaBoardCellStatePlayer1:
      board->numForPlayer1++;
      break;

    case fanoronaBoardCellStatePlayer2:
      board->numForPlayer2++;
      break;

    case fanoronaBoardCellStateInvalid:
      fprintf(stderr,"fanoronaBoardSetCell(): invalid state\n");
      errno=EINVAL;
      return -1;

    default:
      fprintf(stderr,"fanoronaBoardSetCell(): unknown state\n");
      errno=EINVAL;
      return -1;
    }

  switch (board->cells[row][col])
    {
    case fanoronaBoardCellStateEmpty:
      break;

    case fanoronaBoardCellStatePlayer1:
      board->numForPlayer1--;
      break;

    case fanoronaBoardCellStatePlayer2:
      board->numForPlayer2--;
      break;
    }

  board->cells[row][col]=state;

  triggerWatchers(board,col,row,state);

  return 0;
}

FanoronaBoardCellState fanoronaBoardGetCell(const FanoronaBoard* board,
                                            int col,
                                            int row)
{
  if (!board)
    {
      fprintf(stderr,"fanoronaBoardGetCell(): board is NULL\n");
      errno=EFAULT;
      return fanoronaBoardCellStateInvalid;
    }

  if (row<0)
    {
      fprintf(stderr,"fanoronaBoardGetCell(): row<0\n");
      errno=EINVAL;
      return fanoronaBoardCellStateInvalid;
    }

  if (row>=FANORONA_NUM_ROWS)
    {
      fprintf(stderr,"fanoronaBoardGetCell(): row>%d\n",
              FANORONA_NUM_ROWS);
      errno=EINVAL;
      return fanoronaBoardCellStateInvalid;
    }

  if (col<0)
    {
      fprintf(stderr,"fanoronaBoardGetCell(): col<0\n");
      errno=EINVAL;
      return fanoronaBoardCellStateInvalid;
    }

  if (col>=FANORONA_NUM_COLS)
    {
      fprintf(stderr,"fanoronaBoardGetCell(): col>%d\n",
              FANORONA_NUM_COLS);
      errno=EINVAL;
      return fanoronaBoardCellStateInvalid;
    }

  return board->cells[row][col];
}

int fanoronaBoardTestAdjacent(int col1, int row1,
                              int col2, int row2)
{
  if ((col1<0) || (col1>=FANORONA_NUM_COLS))
    return 0;
  if ((row1<0) || (row1>=FANORONA_NUM_ROWS))
    return 0;
  if ((col2<0) || (col2>=FANORONA_NUM_COLS))
    return 0;
  if ((row2<0) || (row2>=FANORONA_NUM_ROWS))
    return 0;

  {
    int dc=col1-col2;
    if ((dc<-1) || (dc>1))
      return 0;
  }
  {
    int dr=row1-row2;
    if ((dr<-1) || (dr>1))
      return 0;
  }

  if (row1==row2)
    return 1;

  if (col1==col2)
    return 1;

  if (row1%2)
    return (col1%2);
  else
    return !(col1%2);
}

int fanoronaBoardIsLegalPos(int col, int row)
{
  if (col<0)
    return 0;
  if (row<0)
    return 0;
  if (col>=FANORONA_NUM_COLS)
    return 0;
  if (row>=FANORONA_NUM_ROWS)
    return 0;

  return 1;
}

FanoronaBoard* fanoronaBoardNew(void)
{
  FanoronaBoard* res=(FanoronaBoard*)malloc(sizeof(FanoronaBoard));
  if (!res)
    {
      int tmp=errno;
      fprintf(stderr,"fanoronaBoardNew(): can't malloc() board: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  if (fanoronaBoardInit(res)<0)
    {
      int tmp=errno;
      free(res);
      fprintf(stderr,"fanoronaBoardNew(): can't init board: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  return res;
}

int fanoronaBoardCopy(FanoronaBoard* dest,
                      const FanoronaBoard* src)
{
  int col,row;

  if (!dest)
    {
      fprintf(stderr,"fanoronaBoardCopy(): dest is NULL\n");
      errno=EFAULT;
      return -1;
    }

  if (!src)
    {
      fprintf(stderr,"fanoronaBoardCopy(): src is NULL\n");
      errno=EFAULT;
      return -1;
    }

  for (col=0; col<FANORONA_NUM_COLS; col++)
    for (row=0; row<FANORONA_NUM_ROWS; row++)
      dest->cells[row][col]=src->cells[row][col];
  dest->numForPlayer1=src->numForPlayer1;
  dest->numForPlayer2=src->numForPlayer2;

  return 0;
}

FanoronaBoard* fanoronaBoardClone(const FanoronaBoard* board)
{
  FanoronaBoard* res=fanoronaBoardNew();
  if (!res)
    {
      int tmp=errno;
      fprintf(stderr,"fanoronaBoardClone(): can't create board: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  if (fanoronaBoardCopy(res,board)<0)
    {
      int tmp=errno;
      fanoronaBoardFree(res);
      fprintf(stderr,"fanoronaBoardClone(): can't copy board: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  return res;
}

void fanoronaBoardFree(FanoronaBoard* board)
{
  if (!board)
    {
      fprintf(stderr,"fanoronaBoardFree(): board is NULL\n");
      errno=EFAULT;
      return;
    }

  while (board->watchers)
    fanoronaBoardDelWatcher(board,board->watchers);

  free(board);
  errno=0;
  return;
}

FanoronaBoard* fanoronaBoardCloneAndSet(FanoronaBoard* board,
                                        int col,
                                        int row,
                                        FanoronaBoardCellState state)
{
  FanoronaBoard* res=fanoronaBoardClone(board);
  if (!res)
    {
      int tmp=errno;
      fprintf(stderr,"fanoronaBoardCloneAndSet(): can't clone board: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  if (fanoronaBoardSetCell(res,col,row,state)<0)
    {
      int tmp=errno;
      fanoronaBoardFree(res);
      fprintf(stderr,"fanoronaBoardCloneAndSet(): can't set cell: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  return res;
}

int fanoronaBoardMove(FanoronaBoard* board,
                      int colFrom,
                      int rowFrom,
                      int colTo,
                      int rowTo,
                      int byApproach,
                      int* numCaptured)
{
  int dc,dr;
  FanoronaBoardCellState self,enemy;

  if (numCaptured)
    (*numCaptured)=0;

  if ((colFrom<0) || (colFrom>=FANORONA_NUM_COLS))
    {
      fprintf(stderr,"fanoronaBoardMove(): colFrom %d out of range\n",
              colFrom);
      return -1;
    }

  if ((rowFrom<0) || (rowFrom>=FANORONA_NUM_ROWS))
    {
      fprintf(stderr,"fanoronaBoardMove(): rowFrom %d out of range\n",
              rowFrom);
      return -1;
    }

  if ((colTo<0) || (colTo>=FANORONA_NUM_COLS))
    {
      fprintf(stderr,"fanoronaBoardMove(): colTo %d out of range\n",
              colTo);
      return -1;
    }

  if ((rowTo<0) || (rowTo>=FANORONA_NUM_ROWS))
    {
      fprintf(stderr,"fanoronaBoardMove(): rowTo %d out of range\n",
              rowTo);
      return -1;
    }

  if (!board)
    {
      fprintf(stderr,"fanoronaBoardMove(): board is null\n");
      return -1;
    }

  if (!fanoronaBoardTestAdjacent(colFrom,rowFrom,colTo,rowTo))
    {
      fprintf(stderr,
              "fanoronaBoardMove(): (%d,%d) is not adjacent to (%d,%d)\n",
              colFrom,rowFrom,colTo,rowTo);
      return -1;
    }

  if (fanoronaBoardGetCell(board,colTo,rowTo)!=fanoronaBoardCellStateEmpty)
    {
      fprintf(stderr,
              "fanoronaBoardMove(): (%d,%d) is not empty\n",
              colTo,rowTo);
      return -1;
    }

  self=fanoronaBoardGetCell(board,colFrom,rowFrom);
  if (self==fanoronaBoardCellStateEmpty)
    {
      fprintf(stderr,
              "fanoronaBoardMove(): (%d,%d) is empty\n",
              colFrom,rowFrom);
      return -1;
    }

  enemy=fanoronaBoardCellOppositeState(self);
  fanoronaBoardSetCell(board,colTo,rowTo,self);
  fanoronaBoardSetCell(board,colFrom,rowFrom,fanoronaBoardCellStateEmpty);

  dc=colTo-colFrom;
  dr=rowTo-rowFrom;

  {
    int col,row;
    if (byApproach)
      {
        col=colTo;
        row=rowTo;
      }
    else
      {
        col=colFrom;
        row=rowFrom;
        dc=-dc;
        dr=-dr;
      }

    col+=dc;
    row+=dr;

    while (   (col<FANORONA_NUM_COLS)
           && (col>=0)
           && (row<FANORONA_NUM_ROWS)
           && (row>=0)
           && (fanoronaBoardGetCell(board,col,row)==enemy)
           )
      {
        fanoronaBoardSetCell(board,col,row,fanoronaBoardCellStateEmpty);
        col+=dc;
        row+=dr;
        if (numCaptured)
          (*numCaptured)++;
      }
  }

  return 0;
}

int fanoronaBoardCopyAndMove(FanoronaBoard* dest,
                             const FanoronaBoard* src,
                             int colFrom,
                             int rowFrom,
                             int colTo,
                             int rowTo,
                             int byApproach,
                             int* numCaptured)
{
  if (!dest)
    {
      fprintf(stderr,"fanoronaBoardCopyAndMove(): dest is NULL\n");
      errno=EFAULT;
      return -1;
    }

  if (!src)
    {
      fprintf(stderr,"fanoronaBoardCopyAndMove(): src is NULL\n");
      errno=EFAULT;
      return -1;
    }

  if (fanoronaBoardCopy(dest,src)<0)
    {
      int tmp=errno;
      fprintf(stderr,"fanoronaBoardCopyAndMove(): can't copy board: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  if (fanoronaBoardMove(dest,colFrom,rowFrom,colTo,rowTo,byApproach,
                        numCaptured)<0)
    {
      int tmp=errno;
      fprintf(stderr,"fanoronaBoardCopyAndMove(): can't make move: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  return 0;
}

static char stateToChar(FanoronaBoardCellState state)
{
  switch (state)
    {
    case fanoronaBoardCellStateEmpty:
      return '.';

    case fanoronaBoardCellStatePlayer1:
      return '1';

    case fanoronaBoardCellStatePlayer2:
      return '2';

    case fanoronaBoardCellStateInvalid:
      return 'I';

    default:
      return '?';
    }
}

void fanoronaBoardPrint(FILE* out,
                        const FanoronaBoard* board)
{
  int col,row;

  if (!board)
    {
      fprintf(stderr,"fanoronaBoardPrint(): board is NULL\n");
      errno=EFAULT;
      return;
    }

  if (!out)
    {
      fprintf(stderr,"fanoronaBoardPrint(): out is NULL\n");
      errno=EFAULT;
      return;
    }

  for (row=0; row<FANORONA_NUM_ROWS; row++)
    {
      for (col=0; col<FANORONA_NUM_COLS; col++)
        {
          if (col>0)
            fprintf(out,"-");
          fprintf(out,"%c",stateToChar(board->cells[row][col]));
        }
      fprintf(out,"\n");
      if (row%2)
        fprintf(out,"|/|\\|/|\\|/|\\|/|\\|\n");
      else
        {
          if ((row+1)<(FANORONA_NUM_ROWS))
            fprintf(out,"|\\|/|\\|/|\\|/|\\|/|\n");
        }
    }
}

FanoronaBoardWatcher
fanoronaBoardAddWatcher(FanoronaBoard* board,
                        FanoronaBoardCellCallback callback,
                        void* context)
{
  FanoronaBoardWatcher res;

  if (!(board && callback))
    {
      errno=EINVAL;
      return 0;
    }

  res=(FanoronaBoardWatcher)malloc(sizeof(struct _FanoronaBoardWatcher));
  if (!res)
    return 0;

  res->callback=callback;
  res->context=context;

  res->next=board->watchers;
  if (res->next)
    res->next->prev=res;

  res->prev=0;
  board->watchers=res;

  return res;
}

void
fanoronaBoardDelWatcher(FanoronaBoard* board,
                        FanoronaBoardWatcher watcher)
{
  if (!(board && watcher))
    return;

  if (watcher->next)
    watcher->next->prev=watcher->prev;
  if (watcher->prev)
    watcher->prev->next=watcher->next;
  else
    board->watchers=watcher->next;
  free(watcher);
}
