/* Copyright (c) 2000 John Stracke <fanorona@thibault.org>
 #
 * This file is part of Fanorona.
 #
 * Fanorona is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any later
 * version. 
 #
 * Fanorona is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details. 
 #
 * You should have received a copy of the GNU General Public License along
 * with program; see the file COPYING. If not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
 */

#include <math.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <sys/param.h>

#include <gtk/gtkmain.h>
#include <gtk/gtksignal.h>

#include <gtk/gtkfanorona.h>

#include <gnome.h>
#include <libgnome/gnome-config.h>

#define FANORONA_DEFAULT_WIDTH 260
#define FANORONA_DEFAULT_HEIGHT 140

enum {
  FANORONA_CLICKED,
  FANORONA_MOVED,
  FANORONA_GAME_OVER,
  FANORONA_PROMPT_FOR_DIRECTION,
  FANORONA_REPORT_WHOSE_MOVE,
  LAST_SIGNAL
};

static gint fanorona_signals[LAST_SIGNAL] = {0,0,0,0,0};

/* Forward declarations */

static void board_changed(FanoronaBoard* board,
                          int col,
                          int row,
                          FanoronaBoardCellState state,
                          void* context);

static void gtk_fanorona_class_init               (GtkFanoronaClass    *klass);
static void gtk_fanorona_init                     (GtkFanorona         *fanorona);
static void gtk_fanorona_destroy                  (GtkObject        *object);
static void gtk_fanorona_realize                  (GtkWidget        *widget);
static void gtk_fanorona_size_request             (GtkWidget      *widget,
					       GtkRequisition *requisition);
static void gtk_fanorona_size_allocate            (GtkWidget     *widget,
					       GtkAllocation *allocation);
static gint gtk_fanorona_expose                   (GtkWidget        *widget,
						GdkEventExpose   *event);
static gint gtk_fanorona_button_press             (GtkWidget        *widget,
						GdkEventButton   *event);
static gint gtk_fanorona_button_release           (GtkWidget        *widget,
						GdkEventButton   *event);

static void clear_all_possibles(GtkFanorona* fanorona,
                                gboolean invalidate);
static void clear_all_useds(GtkFanorona* fanorona,
                            gboolean invalidate);

static void clicked(GtkWidget* widget,
                    gint col,
                    gint row,
                    gpointer data);
static void moved(GtkFanorona* fanorona,
                  const GtkFanoronaMove* move,
                  gpointer data);

static void invalidate_all(GtkFanorona* fanorona);
static void _gtk_fanorona_set_color_empty(GtkFanorona* fanorona,
                                          const char* empty);
static void _gtk_fanorona_set_color_player1(GtkFanorona* fanorona,
                                            const char* player1);
static void _gtk_fanorona_set_color_player2(GtkFanorona* fanorona,
                                            const char* player2);
static void _gtk_fanorona_set_color_borders(GtkFanorona* fanorona,
                                            const char* borders);
static void _gtk_fanorona_set_color_pressed(GtkFanorona* fanorona,
                                            const char* pressed);
static void _gtk_fanorona_set_color_possible(GtkFanorona* fanorona,
                                             const char* possible);

/* Local data */

static GtkWidgetClass *parent_class = NULL;

guint
gtk_fanorona_get_type ()
{
  static guint fanorona_type = 0;

  if (!fanorona_type)
    {
      GtkTypeInfo fanorona_info =
      {
	"GtkFanorona",
	sizeof (GtkFanorona),
	sizeof (GtkFanoronaClass),
	(GtkClassInitFunc) gtk_fanorona_class_init,
	(GtkObjectInitFunc) gtk_fanorona_init,
	(GtkArgSetFunc) NULL,
	(GtkArgGetFunc) NULL,
      };

      fanorona_type = gtk_type_unique (gtk_widget_get_type (), &fanorona_info);
    }

  return fanorona_type;
}

static void
gtk_fanorona_class_init (GtkFanoronaClass *class)
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;

  object_class = (GtkObjectClass*) class;
  widget_class = (GtkWidgetClass*) class;

  parent_class = gtk_type_class (gtk_widget_get_type ());

  object_class->destroy = gtk_fanorona_destroy;

  widget_class->realize = gtk_fanorona_realize;
  widget_class->expose_event = gtk_fanorona_expose;
  widget_class->size_request = gtk_fanorona_size_request;
  widget_class->size_allocate = gtk_fanorona_size_allocate;
  widget_class->button_press_event = gtk_fanorona_button_press;
  widget_class->button_release_event = gtk_fanorona_button_release;

  class->clicked=NULL;
  class->moved=NULL;
  class->game_over=NULL;
  class->prompt_for_direction=NULL;
  class->report_whose_move=NULL;

  fanorona_signals[FANORONA_CLICKED]
    =gtk_signal_new("fanorona_clicked",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET(GtkFanoronaClass,
                                      clicked),
                    gtk_marshal_NONE__INT_INT, GTK_TYPE_NONE,
                    2,GTK_TYPE_INT,GTK_TYPE_INT);
  fanorona_signals[FANORONA_MOVED]
    =gtk_signal_new("fanorona_moved",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET(GtkFanoronaClass,
                                      moved),
                    gtk_marshal_NONE__POINTER, GTK_TYPE_NONE,
                    1,GTK_TYPE_POINTER);
  fanorona_signals[FANORONA_GAME_OVER]
    =gtk_signal_new("fanorona_game_over",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET(GtkFanoronaClass,
                                      game_over),
                    gtk_marshal_NONE__BOOL, GTK_TYPE_NONE,
                    1,GTK_TYPE_BOOL);
  fanorona_signals[FANORONA_PROMPT_FOR_DIRECTION]
    =gtk_signal_new("fanorona_prompt_for_direction",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET(GtkFanoronaClass,
                                      prompt_for_direction),
                    gtk_marshal_NONE__NONE, GTK_TYPE_NONE,
                    0);
  fanorona_signals[FANORONA_REPORT_WHOSE_MOVE]
    =gtk_signal_new("fanorona_report_whose_move",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET(GtkFanoronaClass,
                                      report_whose_move),
                    gtk_marshal_NONE__BOOL, GTK_TYPE_NONE,
                    1,GTK_TYPE_BOOL);
  gtk_object_class_add_signals (object_class,
                                fanorona_signals,
                                LAST_SIGNAL);
}

static void
gtk_fanorona_init(GtkFanorona *fanorona)
{
  fanorona->board = NULL;
  fanorona->boardWasCreated=0;
  fanorona->watcher = NULL;
  fanorona->cellsize=0;
  fanorona->xmargin=0;
  fanorona->ymargin=0;
  fanorona->pressedIn.col=-1;
  fanorona->pressedIn.row=-1;

  fanorona->gcs.empty=0;
  fanorona->gcs.player1=0;
  fanorona->gcs.player2=0;
  fanorona->gcs.borders=0;
  fanorona->gcs.pressed=0;
  fanorona->gcs.possible=0;

  fanorona->delay=500;

  gtk_fanorona_set_colors(fanorona,
                          "white","red","green","black","white","yellow",
                          0);
  gtk_fanorona_set_cur_player(fanorona,fanoronaBoardCellStatePlayer1);
  gtk_fanorona_set_click_validator(fanorona,0,0);

  fanorona->savedBoards=0;
  fanorona->latestBoard=0;

  fanorona->moveInProgress.colFrom=-1;
  fanorona->moveInProgress.rowFrom=-1;
  fanorona->ambiguousMoveInProgress=FALSE;
  /*fanorona->firstMoveOfTurn=FALSE;*/

  clear_all_possibles(fanorona,FALSE);
  clear_all_useds(fanorona,FALSE);
}

static int saveExternState(GtkFanorona* fanorona,
                           void** data,
                           int* len)
{
  if (!fanorona)
    {
      fprintf(stderr,"saveExternState(): fanorona is NULL\n");
      errno=EFAULT;
      return -1;
    }

  if (!data)
    {
      fprintf(stderr,"saveExternState(): data is NULL\n");
      errno=EFAULT;
      return -1;
    }

  if (!len)
    {
      fprintf(stderr,"saveExternState(): len is NULL\n");
      errno=EFAULT;
      return -1;
    }

  if (fanorona->stateMethods.get)
    {
      (*data)=(fanorona->stateMethods.get)(fanorona->stateMethods.context,
                                           len);
      if (!(*data))
        {
          int tmp=errno;
          fprintf(stderr,"saveExternState(): get method failed: %s\n",
                  strerror(tmp));
          errno=tmp;
          return -1;
        }

      return 0;
    }
  else
    {
      (*data)=0;
      (*len)=0;

      return 0;
    }
}

static int restoreExternState(GtkFanorona* fanorona,
                              void* data,
                              int len)
{
  if (!fanorona)
    {
      fprintf(stderr,"restoreExternState(): fanorona is NULL\n");
      errno=EFAULT;
      return -1;
    }

  if (fanorona->stateMethods.restore)
    {
      if (((fanorona->stateMethods.restore)(fanorona->stateMethods.context,
                                            data,
                                            len)
           )
          <0
          )
        {
          int tmp=errno;
          fprintf(stderr,"restoreExternState(): restore method failed: %s\n",
                  strerror(tmp));
          errno=tmp;
          return -1;
        }
    }

  return 0;
}

static int pushCurBoard(GtkFanorona* fanorona)
{
  GtkFanorona_SavedBoard* saved;

  if (!fanorona)
    {
      fprintf(stderr,"pushCurBoard(): fanorona is NULL\n");
      errno=EFAULT;
      return -1;
    }

  if (fanorona->latestBoard)
    {
      GtkFanorona_SavedBoard* cur=fanorona->latestBoard->prev;
      while (cur)
        {
          GtkFanorona_SavedBoard* prev=cur->prev;
          free(cur);
          cur=prev;
        }
      fanorona->savedBoards=fanorona->latestBoard;
      fanorona->latestBoard=0;
    }

  saved=(GtkFanorona_SavedBoard*)malloc(sizeof(GtkFanorona_SavedBoard));
  if (!saved)
    {
      int tmp=errno;
      fprintf(stderr,"pushCurBoard(): can't malloc() saved board: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  if (fanoronaBoardInit(&(saved->board))<0)
    {
      int tmp=errno;
      fprintf(stderr,"pushCurBoard(): can't init saved board: %s\n",
              strerror(tmp));
      free(saved);
      errno=tmp;
      return -1;
    }

  if (fanoronaBoardCopy(&(saved->board),fanorona->board)<0)
    {
      int tmp=errno;
      fprintf(stderr,"pushCurBoard(): can't copy current board: %s\n",
              strerror(tmp));
      free(saved);
      errno=tmp;
      return -1;
    }

  if (saveExternState(fanorona,
                      &(saved->externState.data),
                      &(saved->externState.len))<0)
    {
      int tmp=errno;
      fprintf(stderr,"pushCurBoard(): can't save extern state: %s\n",
              strerror(tmp));
      free(saved);
      errno=tmp;
      return -1;
    }

  saved->pressedIn.col=fanorona->pressedIn.col;
  saved->pressedIn.row=fanorona->pressedIn.row;
  saved->curPlayer=fanorona->curPlayer;
  saved->moveInProgress.colFrom=fanorona->moveInProgress.colFrom;
  saved->moveInProgress.rowFrom=fanorona->moveInProgress.rowFrom;
  saved->ambiguousMoveInProgress=fanorona->ambiguousMoveInProgress;
  saved->ambiguousMove=fanorona->ambiguousMove;

  {
    int col,row;
    for (col=0; col<FANORONA_NUM_COLS; col++)
      for (row=0; row<FANORONA_NUM_ROWS; row++)
        {
          saved->possibles[col][row]=fanorona->possibles[col][row];
          saved->useds[col][row]=fanorona->useds[col][row];
        }
  }

  saved->next=fanorona->savedBoards;
  saved->prev=0;
  if (fanorona->savedBoards)
    {
      fanorona->savedBoards->prev=saved;
      saved->turnNum=(fanorona->savedBoards->turnNum)+1;
    }
  else
    saved->turnNum=1;

  fanorona->savedBoards=saved;

  return 0;
}

static int restoreSavedBoard(GtkFanorona* fanorona,
                             GtkFanorona_SavedBoard* saved)
{
  if (saved)
    {
      if (fanoronaBoardCopy(fanorona->board,&(saved->board))<0)
        {
          int tmp=errno;
          fprintf(stderr,"restoreSavedBoard(): can't copy saved board: %s\n",
                  strerror(tmp));
          free(saved);
          errno=tmp;
          return -1;
        }
  
      fanorona->pressedIn.col=saved->pressedIn.col;
      fanorona->pressedIn.row=saved->pressedIn.row;
      fanorona->curPlayer=saved->curPlayer;
      fanorona->moveInProgress.colFrom=saved->moveInProgress.colFrom;
      fanorona->moveInProgress.rowFrom=saved->moveInProgress.rowFrom;
      fanorona->ambiguousMoveInProgress=saved->ambiguousMoveInProgress;
      fanorona->ambiguousMove=saved->ambiguousMove;
  
      {
        int col,row;
        for (col=0; col<FANORONA_NUM_COLS; col++)
          for (row=0; row<FANORONA_NUM_ROWS; row++)
            {
              fanorona->possibles[col][row]=saved->possibles[col][row];
              fanorona->useds[col][row]=saved->useds[col][row];
            }
      }
  
      if (restoreExternState(fanorona,
                             saved->externState.data,
                             saved->externState.len)
          <0
          )
        {
          int tmp=errno;
          fprintf(stderr,"restoreCurBoard(): can't restore extern state: %s\n",
                  strerror(tmp));
          free(saved);
          errno=tmp;
          return -1;
        }
    }
  else
    {
      if (fanoronaBoardSetup(fanorona->board)<0)
        {
          int tmp=errno;
          fprintf(stderr,"restoreSavedBoard(): can't set up board: %s\n",
                  strerror(tmp));
          free(saved);
          errno=tmp;
          return -1;
        }

      fanorona->pressedIn.col=-1;
      fanorona->pressedIn.row=-1;
      gtk_fanorona_set_cur_player(fanorona,fanoronaBoardCellStatePlayer1);
      fanorona->moveInProgress.colFrom=-1;
      fanorona->moveInProgress.rowFrom=-1;
      fanorona->ambiguousMoveInProgress=0;

      clear_all_possibles(fanorona,FALSE);
      clear_all_useds(fanorona,FALSE);
    }

  {
    GtkWidget* widget=GTK_WIDGET(fanorona);

    GdkEventExpose event;
    event.type=GDK_EXPOSE;
    event.window=widget->window;
    event.send_event=1;
    event.area.x=0;
    event.area.y=0;
    event.area.width=widget->allocation.width;
    event.area.height=widget->allocation.height;
    event.count=0;

    gdk_event_put((GdkEvent*)&event);
  }
  
  return 0;
}

static int popSavedBoard(GtkFanorona* fanorona)
{
  GtkFanorona_SavedBoard* saved;

  if (!fanorona)
    {
      fprintf(stderr,"popSavedBoard(): fanorona is NULL\n");
      errno=EFAULT;
      return -1;
    }

  if (fanorona->latestBoard)
    {
      saved=fanorona->latestBoard;
      if (fanorona->latestBoard->next)
        fanorona->latestBoard=fanorona->latestBoard->next;
    }
  else
    {
      saved=fanorona->savedBoards;
      fanorona->latestBoard=saved->next;
    }

  if (restoreSavedBoard(fanorona,saved)<0)
    {
      int tmp=errno;
      fprintf(stderr,"popSavedBoard(): can't restore state: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  /*fanorona->firstMoveOfTurn=1;*/
  return 0;
}

static int clearSavedBoards(GtkFanorona* fanorona)
{
  GtkFanorona_SavedBoard* saved;

  if (!fanorona)
    {
      fprintf(stderr,"clearSavedBoards(): fanorona is NULL\n");
      errno=EFAULT;
      return -1;
    }

  while (fanorona->savedBoards)
    {
      GtkFanorona_SavedBoard* cur=fanorona->savedBoards;
      fanorona->savedBoards=cur->next;
      free(cur);
    }

  fanorona->latestBoard=0;

  return 0;
}

static gboolean clickIsValid(GtkFanorona* fanorona,
                             int col,
                             int row)
{
  if (fanorona->clickValidator.function)
    return
      (fanorona->clickValidator.function)(fanorona,col,row,
                                          fanorona->clickValidator.context);
  else
    return 1;
}

GtkWidget*
gtk_fanorona_new (FanoronaBoard* board)
{
  GtkFanorona *fanorona;

  fanorona = gtk_type_new (gtk_fanorona_get_type ());

  if (!board)
    {
      board = fanoronaBoardNew();
      fanorona->boardWasCreated=1;
    }
  else
    fanorona->boardWasCreated=0;

  fanoronaBoardSetup(board);

  gtk_fanorona_set_board (fanorona, board);

  return GTK_WIDGET (fanorona);
}

static void
gtk_fanorona_destroy (GtkObject *object)
{
  GtkFanorona *fanorona;

  g_return_if_fail (object != NULL);
  g_return_if_fail (GTK_IS_FANORONA (object));

  fanorona = GTK_FANORONA (object);

  if (fanorona->boardWasCreated)
    fanoronaBoardFree(fanorona->board);

  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

FanoronaBoard*
gtk_fanorona_get_board (GtkFanorona *fanorona)
{
  g_return_val_if_fail (fanorona != NULL, NULL);
  g_return_val_if_fail (GTK_IS_FANORONA (fanorona), NULL);

  return fanorona->board;
}

void
gtk_fanorona_set_board (GtkFanorona      *fanorona,
                        FanoronaBoard* board)
{
  g_return_if_fail (fanorona != NULL);
  g_return_if_fail (GTK_IS_FANORONA (fanorona));

  if (fanorona->board)
    {
      if (fanorona->boardWasCreated)
        fanoronaBoardFree(fanorona->board);
      else
        fanoronaBoardDelWatcher(fanorona->board,
                                fanorona->watcher);
    }

  fanorona->board = board;
  fanorona->boardWasCreated=0;

  fanorona->watcher=fanoronaBoardAddWatcher(fanorona->board,
                                            board_changed,
                                            fanorona);
}

static void dumpStyleColors(GdkColor* first,
                            const char* tag)
{
  int i;
  printf(" %s colors:\n",tag);
  for (i=0; i<5; i++)
    {
      GdkColor* cur=first+i;
      printf("  pixel #%ul; RGB (%d,%d,%d)\n",
             (unsigned long)(cur->pixel),
             (int)(cur->red),
             (int)(cur->green),
             (int)(cur->blue));
    }
}

static void dumpStyle(GtkStyle* style,
                      const char* where)
{
  printf("dumpStyle(%s)\n",where);
  dumpStyleColors(style->fg,"fg");
  dumpStyleColors(style->bg,"bg");
  dumpStyleColors(style->light,"light");
  dumpStyleColors(style->dark,"dark");
  dumpStyleColors(style->mid,"mid");
  dumpStyleColors(style->text,"text");
  dumpStyleColors(style->base,"base");
}

static void
gtk_fanorona_realize (GtkWidget *widget)
{
  GdkWindowAttr attributes;
  gint attributes_mask;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_FANORONA (widget));

  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

  /*dumpStyle(widget->style,"realize");*/

  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.event_mask = gtk_widget_get_events (widget) | 
    GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK | 
    GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK |
    GDK_POINTER_MOTION_HINT_MASK;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);

  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
  widget->window = gdk_window_new (widget->parent->window, &attributes, attributes_mask);

  widget->style = gtk_style_attach (widget->style, widget->window);

  gdk_window_set_user_data (widget->window, widget);

  gtk_style_set_background (widget->style, widget->window, GTK_STATE_ACTIVE);
}

static void 
gtk_fanorona_size_request (GtkWidget      *widget,
		       GtkRequisition *requisition)
{
  requisition->width = FANORONA_DEFAULT_WIDTH;
  requisition->height = FANORONA_DEFAULT_HEIGHT;
}

static void
gtk_fanorona_size_allocate (GtkWidget     *widget,
                            GtkAllocation *allocation)
{
  GtkFanorona *fanorona;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_FANORONA (widget));
  g_return_if_fail (allocation != NULL);

  widget->allocation = *allocation;
  fanorona = GTK_FANORONA (widget);

  if (GTK_WIDGET_REALIZED (widget))
    {
      gdk_window_move_resize (widget->window,
			      allocation->x, allocation->y,
			      allocation->width, allocation->height);

    }

  {
    int xs=(allocation->width)/(FANORONA_NUM_COLS*3-1);
    int ys=(allocation->height)/(FANORONA_NUM_ROWS*3-1);

    fanorona->cellsize=MIN(xs,ys);
  }

  fanorona->xmargin=((allocation->width)
                     -((fanorona->cellsize)*(FANORONA_NUM_COLS*3-1)))/2;
  fanorona->ymargin=((allocation->height)
                     -((fanorona->cellsize)*(FANORONA_NUM_ROWS*3-1)))/2;
}

static void
draw_connection(GtkFanorona* fanorona,
                GdkGC* gc,
                int col1, int row1,
                int col2, int row2)
{
  GtkWidget* widget=GTK_WIDGET(fanorona);
  int x1,y1,x2,y2;

  if (!fanoronaBoardTestAdjacent(col1,row1,col2,row2))
    return;

  x1=col1*3+1;
  y1=row1*3+1;
  x2=col2*3+1;
  y2=row2*3+1;

  if (x2>x1)
    {
      x1++;
      x2--;
    }
  else
    {
      if (x2<x1)
        {
          x1--;
          x2++;
        }
    }

  if (y2>y1)
    {
      y1++;
      y2--;
    }
  else
    {
      if (y2<y1)
        {
          y1--;
          y2++;
        }
    }

  x1*=(fanorona->cellsize);
  y1*=(fanorona->cellsize);
  x2*=(fanorona->cellsize);
  y2*=(fanorona->cellsize);

  x1+=(fanorona->xmargin);
  y1+=(fanorona->ymargin);
  x2+=(fanorona->xmargin);
  y2+=(fanorona->ymargin);

  {
    GdkGC* g=gdk_gc_new(fanorona->widget.window);
    gdk_gc_copy(g,gc);
    gdk_gc_set_line_attributes(g,
                               3,
                               GDK_LINE_SOLID,
                               GDK_CAP_NOT_LAST,
                               GDK_JOIN_MITER);
    gdk_draw_line(widget->window,
                  g,
                  x1,y1,x2,y2);
  }
}                

static gboolean areaIsExposed(GdkEventExpose *event,
                              int x,
                              int y,
                              int w,
                              int h)
{
  GdkRectangle xywh,tst;

  xywh.x=x;
  xywh.y=y;
  xywh.width=w;
  xywh.height=h;

  return gdk_rectangle_intersect(&(event->area),&xywh,&tst);
}

static gint
gtk_fanorona_expose (GtkWidget      *widget,
                     GdkEventExpose *event)
{
  GtkFanorona *fanorona;
  int row,col;
  static int toggled=0;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_FANORONA (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  fanorona = GTK_FANORONA (widget);

  /*dumpStyle(widget->style,"expose");*/

  if (!(fanorona->gcs.empty))
    gtk_fanorona_set_colors(fanorona,
                            0,0,0,0,0,0,
                            0);

  gtk_style_set_background(widget->style,
                           widget->window,
                           GTK_WIDGET_STATE(widget));

  for (row=0; row<FANORONA_NUM_ROWS; row++)
    {
      for (col=0; col<FANORONA_NUM_COLS; col++)
        {
          int x=(fanorona->cellsize)*col*3+(fanorona->xmargin);
          int y=(fanorona->cellsize)*row*3+(fanorona->ymargin);
          int widthOrHeight=(fanorona->cellsize)*2;

          GdkGC* bkg_gc=fanorona->gcs.borders;

          int isPressed=(   (col==fanorona->pressedIn.col)
                            && (row==fanorona->pressedIn.row)
                            );
          GdkGC* gc;

          if (isPressed)
            bkg_gc=fanorona->gcs.pressed;
          else
            {
              if (gtk_fanorona_get_possible(fanorona,col,row))
                bkg_gc=fanorona->gcs.possible;
            }

          {
            int w=(fanorona->cellsize)*4;
            int h=(fanorona->cellsize)*3;
            if (!areaIsExposed(event,
                               x-(fanorona->cellsize),y,
                               w,h))
              continue;
          }

          gdk_draw_rectangle(widget->window,
                             bkg_gc,
                             TRUE,
                             x,y,widthOrHeight,widthOrHeight);

          draw_connection(fanorona,fanorona->gcs.borders,col,row,col,row+1);
          draw_connection(fanorona,fanorona->gcs.borders,col,row,col+1,row);
          draw_connection(fanorona,fanorona->gcs.borders,col,row,col+1,row+1);
          draw_connection(fanorona,fanorona->gcs.borders,col,row,col-1,row+1);

          switch (fanoronaBoardGetCell(fanorona->board,col,row))
            {
            case fanoronaBoardCellStatePlayer1:
              gc=fanorona->gcs.player1;
              break;

            case fanoronaBoardCellStatePlayer2:
              gc=fanorona->gcs.player2;
              break;

            case fanoronaBoardCellStateEmpty:
            default:
              gc=fanorona->gcs.empty;
              break;
            }

          gdk_draw_rectangle(widget->window,
                             gc,
                             TRUE,
                             x+2,y+2,widthOrHeight-4,widthOrHeight-4);

          if (gtk_fanorona_get_used(fanorona,col,row))
            {
              if ((fanorona->curPlayer)==fanoronaBoardCellStatePlayer1)
                gc=fanorona->gcs.player1;
              else
                gc=fanorona->gcs.player2;

              gdk_draw_line(widget->window,
                            gc,
                            x+1,y+1,
                            x+widthOrHeight-1,y+widthOrHeight-1);
              gdk_draw_line(widget->window,
                            gc,
                            x+widthOrHeight-1,y+1,
                            x+1,y+widthOrHeight-1);
            }
        }
    }

  return FALSE;
}

static int xyToColRow(GtkFanorona* fanorona,
                      int x, int y,
                      int* col, int* row)
{
  int c,r;
  int size3=(fanorona->cellsize*3);
  int size2=(fanorona->cellsize*2);

  x-=(fanorona->xmargin);
  y-=(fanorona->ymargin);

  c=x/size3;
  r=y/size3;
  
  if (x>(c*size3+size2))
    return -1;
  if (y>(r*size3+size2))
    return -1;

  (*col)=c;
  (*row)=r;
  return 0;
}

static gint
gtk_fanorona_button_press (GtkWidget      *widget,
                           GdkEventButton *event)
{
  GtkFanorona *fanorona;
  int col,row;
  FanoronaBoardCellState stateNeeded;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_FANORONA (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  fanorona = GTK_FANORONA (widget);

  if (xyToColRow(fanorona,event->x,event->y,&col,&row)<0)
    {
      printf("Pressed outside of any cell\n");
      fanorona->pressedIn.col=-1;
      fanorona->pressedIn.row=-1;
      return FALSE;
    }

  if (!clickIsValid(fanorona,col,row))
    {
      /*printf("Invalid click\n");*/
      fanorona->pressedIn.col=-1;
      fanorona->pressedIn.row=-1;
      return FALSE;
    }

  /*printf("Pressed in (%d,%d)\n",col,row);*/

  fanorona->pressedIn.col=col;
  fanorona->pressedIn.row=row;

  gtk_fanorona_invalidate_cell(fanorona,col,row);

  return FALSE;
}

static gint
gtk_fanorona_button_release (GtkWidget      *widget,
			  GdkEventButton *event)
{
  GtkFanorona *fanorona;
  int col,row,pressCol,pressRow;
  gboolean fire;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_FANORONA (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  fanorona = GTK_FANORONA (widget);

  pressCol=fanorona->pressedIn.col;
  pressRow=fanorona->pressedIn.row;

  fanorona->pressedIn.col=-1;
  fanorona->pressedIn.row=-1;

  gtk_fanorona_invalidate_cell(fanorona,pressCol,pressRow);

  if (xyToColRow(fanorona,event->x,event->y,&col,&row)<0)
    return FALSE;

  fire=((pressCol==col) && (pressRow==row));

  /*printf("Released in (%d,%d): %sfire\n",
         col,row,(fire ? "" : "do not "));*/

  if (fire)
    gtk_signal_emit(GTK_OBJECT(fanorona),
                    fanorona_signals[FANORONA_CLICKED],
                    (gint)col,(gint)row);

  return FALSE;
}

void gtk_fanorona_invalidate_cell(GtkFanorona* fanorona,
                                  int col,
                                  int row)
{
  GtkWidget* widget=GTK_WIDGET(fanorona);

  GdkEventExpose event;
  event.type=GDK_EXPOSE;
  event.window=widget->window;
  event.send_event=1;
  event.area.x=(fanorona->cellsize)*col*3+(fanorona->xmargin);
  event.area.y=(fanorona->cellsize)*row*3+(fanorona->ymargin);
  event.area.width=event.area.height=(fanorona->cellsize)*2;
  event.count=0;

  /*printf("Exposing (%d,%d)x(%d,%d)\n",
         event.area.x,
         event.area.y,
         event.area.width,
         event.area.height);*/

  gdk_event_put((GdkEvent*)&event);
}

static void board_changed(FanoronaBoard* board,
                          int col,
                          int row,
                          FanoronaBoardCellState state,
                          void* context)
{
  GtkFanorona* fanorona=(GtkFanorona*)context;
  gtk_fanorona_invalidate_cell(fanorona,col,row);

  {
    int numForPlayer1,numForPlayer2;

    fanoronaBoardCountSides(board,
                            &numForPlayer1,
                            &numForPlayer2);
    if ((!numForPlayer1) || (!numForPlayer2))
      gtk_signal_emit(GTK_OBJECT(fanorona),
                      fanorona_signals[FANORONA_GAME_OVER],
                      (numForPlayer1>0));
  }
}

static void _gtk_fanorona_set_color_empty(GtkFanorona* fanorona,
                                          const char* empty)
{
  GtkWidget* widget=GTK_WIDGET(fanorona);
  GdkColormap* cmap;

  if (empty)
    {
      if (*empty)
        gdk_color_parse(empty,&(fanorona->colors.empty));
      else
        fanorona->colors.empty=widget->style->light[GTK_WIDGET_STATE(widget)];
    }

  if (!(widget->window))
    return;
  cmap=gdk_window_get_colormap(widget->window);

  if (!(fanorona->gcs.empty))
    {
      fanorona->gcs.empty=gdk_gc_new(widget->window);
      gdk_gc_copy(fanorona->gcs.empty,widget->style->black_gc);
    }

  gdk_colormap_alloc_color(cmap,&(fanorona->colors.empty),FALSE,TRUE);
  gdk_gc_set_foreground(fanorona->gcs.empty,&(fanorona->colors.empty));
}

static void _gtk_fanorona_set_color_player1(GtkFanorona* fanorona,
                                            const char* player1)
{
  GtkWidget* widget=GTK_WIDGET(fanorona);
  GdkColormap* cmap;

  if (player1)
    {
      if (*player1)
        gdk_color_parse(player1,&(fanorona->colors.player1));
      else
        fanorona->colors.player1=widget->style->mid[GTK_WIDGET_STATE(widget)];
    }

  if (!(widget->window))
    return;
  cmap=gdk_window_get_colormap(widget->window);

  if (!(fanorona->gcs.player1))
    {
      fanorona->gcs.player1=gdk_gc_new(widget->window);
      gdk_gc_copy(fanorona->gcs.player1,widget->style->black_gc);
      gdk_gc_set_line_attributes(fanorona->gcs.player1,
                                 3,
                                 GDK_LINE_SOLID,
                                 GDK_CAP_NOT_LAST,
                                 GDK_JOIN_MITER);
    }

  gdk_colormap_alloc_color(cmap,&(fanorona->colors.player1),FALSE,TRUE);
  gdk_gc_set_foreground(fanorona->gcs.player1,&(fanorona->colors.player1));
}

static void _gtk_fanorona_set_color_player2(GtkFanorona* fanorona,
                                            const char* player2)
{
  GtkWidget* widget=GTK_WIDGET(fanorona);
  GdkColormap* cmap;

  if (player2)
    {
      if (*player2)
        gdk_color_parse(player2,&(fanorona->colors.player2));
      else
        fanorona->colors.player2=widget->style->dark[GTK_WIDGET_STATE(widget)];
    }

  if (!(widget->window))
    return;
  cmap=gdk_window_get_colormap(widget->window);

  if (!(fanorona->gcs.player2))
    {
      fanorona->gcs.player2=gdk_gc_new(widget->window);
      gdk_gc_copy(fanorona->gcs.player2,widget->style->black_gc);
      gdk_gc_set_line_attributes(fanorona->gcs.player2,
                                 3,
                                 GDK_LINE_SOLID,
                                 GDK_CAP_NOT_LAST,
                                 GDK_JOIN_MITER);
    }

  gdk_colormap_alloc_color(cmap,&(fanorona->colors.player2),FALSE,TRUE);
  gdk_gc_set_foreground(fanorona->gcs.player2,&(fanorona->colors.player2));
}

static void _gtk_fanorona_set_color_borders(GtkFanorona* fanorona,
                                            const char* borders)
{
  GtkWidget* widget=GTK_WIDGET(fanorona);
  GdkColormap* cmap;

  if (borders)
    {
      if (*borders)
        gdk_color_parse(borders,&(fanorona->colors.borders));
      else
        fanorona->colors.borders=widget->style->fg[GTK_WIDGET_STATE(widget)];
    }

  if (!(widget->window))
    return;
  cmap=gdk_window_get_colormap(widget->window);

  if (!(fanorona->gcs.borders))
    {
      fanorona->gcs.borders=gdk_gc_new(widget->window);
      gdk_gc_copy(fanorona->gcs.borders,widget->style->black_gc);
    }

  gdk_colormap_alloc_color(cmap,&(fanorona->colors.borders),FALSE,TRUE);

  gdk_gc_set_foreground(fanorona->gcs.borders,&(fanorona->colors.borders));
}

static void _gtk_fanorona_set_color_pressed(GtkFanorona* fanorona,
                                            const char* pressed)
{
  GtkWidget* widget=GTK_WIDGET(fanorona);
  GdkColormap* cmap;

  if (pressed)
    {
      if (*pressed)
        gdk_color_parse(pressed,&(fanorona->colors.pressed));
      else
        fanorona->colors.pressed=widget->style->light[GTK_WIDGET_STATE(widget)];
    }

  if (!(widget->window))
    return;
  cmap=gdk_window_get_colormap(widget->window);

  if (!(fanorona->gcs.pressed))
    {
      fanorona->gcs.pressed=gdk_gc_new(widget->window);
      gdk_gc_copy(fanorona->gcs.pressed,widget->style->black_gc);
    }
  gdk_colormap_alloc_color(cmap,&(fanorona->colors.pressed),FALSE,TRUE);
  gdk_gc_set_foreground(fanorona->gcs.pressed,&(fanorona->colors.pressed));
}

static void _gtk_fanorona_set_color_possible(GtkFanorona* fanorona,
                                             const char* possible)
{
  GtkWidget* widget=GTK_WIDGET(fanorona);
  GdkColormap* cmap;

  if (possible)
    {
      if (*possible)
        gdk_color_parse(possible,&(fanorona->colors.possible));
      else
        fanorona->colors.possible=widget->style->base[GTK_WIDGET_STATE(widget)];
    }


  if (!(widget->window))
    return;
  cmap=gdk_window_get_colormap(widget->window);

  if (!(fanorona->gcs.possible))
    {
      fanorona->gcs.possible=gdk_gc_new(widget->window);
      gdk_gc_copy(fanorona->gcs.possible,widget->style->black_gc);
    }
  gdk_colormap_alloc_color(cmap,&(fanorona->colors.possible),FALSE,TRUE);
  gdk_gc_set_foreground(fanorona->gcs.possible,&(fanorona->colors.possible));
}

void gtk_fanorona_set_color_empty(GtkFanorona* fanorona,
                                  const char* empty)
{
  _gtk_fanorona_set_color_empty(fanorona,empty);
  invalidate_all(fanorona);
}

void gtk_fanorona_set_color_player1(GtkFanorona* fanorona,
                                    const char* player1)
{
  _gtk_fanorona_set_color_player1(fanorona,player1);
  invalidate_all(fanorona);
}

void gtk_fanorona_set_color_player2(GtkFanorona* fanorona,
                                    const char* player2)
{
  _gtk_fanorona_set_color_player2(fanorona,player2);
  invalidate_all(fanorona);
}

void gtk_fanorona_set_color_borders(GtkFanorona* fanorona,
                                    const char* borders)
{
  _gtk_fanorona_set_color_borders(fanorona,borders);
  invalidate_all(fanorona);
}

void gtk_fanorona_set_color_pressed(GtkFanorona* fanorona,
                                    const char* pressed)
{
  _gtk_fanorona_set_color_pressed(fanorona,pressed);
  invalidate_all(fanorona);
}

void gtk_fanorona_set_color_possible(GtkFanorona* fanorona,
                                     const char* possible)
{
  _gtk_fanorona_set_color_possible(fanorona,possible);
  invalidate_all(fanorona);
}

void gtk_fanorona_set_colors(GtkFanorona* fanorona,
                             const char* empty,
                             const char* player1,
                             const char* player2,
                             const char* borders,
                             const char* pressed,
                             const char* possible,
                             gboolean invalidate)
{
  _gtk_fanorona_set_color_empty(fanorona,empty);
  _gtk_fanorona_set_color_player1(fanorona,player1);
  _gtk_fanorona_set_color_player2(fanorona,player2);
  _gtk_fanorona_set_color_borders(fanorona,borders);
  _gtk_fanorona_set_color_pressed(fanorona,pressed);
  _gtk_fanorona_set_color_possible(fanorona,possible);

  if (invalidate)
    invalidate_all(fanorona);
}

void gtk_fanorona_set_cur_player(GtkFanorona* fanorona,
                                 FanoronaBoardCellState state)
{
  fanorona->curPlayer=state;
}

FanoronaBoardCellState
gtk_fanorona_get_cur_player(GtkFanorona* fanorona)
{
  return fanorona->curPlayer;
}

static int safeTurnNum(GtkFanorona_SavedBoard* saved)
{
  return saved ? saved->turnNum : -1;
}

void gtk_fanorona_undo(GtkFanorona* fanorona)
{
  /*printf("gtk_fanorona_undo(): before: (savedBoards,latestBoard)==(%d,%d)\n",
         safeTurnNum(fanorona->savedBoards),
         safeTurnNum(fanorona->latestBoard));*/
  popSavedBoard(fanorona);
  /*printf("gtk_fanorona_undo(): after: (savedBoards,latestBoard)==(%d,%d)\n\n",
         safeTurnNum(fanorona->savedBoards),
         safeTurnNum(fanorona->latestBoard));*/
}

void gtk_fanorona_redo(GtkFanorona* fanorona)
{
  /*printf("gtk_fanorona_redo(): before: (savedBoards,latestBoard)==(%d,%d)\n",
         safeTurnNum(fanorona->savedBoards),
         safeTurnNum(fanorona->latestBoard));*/
  if (fanorona->latestBoard)
    {
      if (fanorona->latestBoard->prev)
        fanorona->latestBoard=fanorona->latestBoard->prev;
    }
  else
    fanorona->latestBoard=fanorona->savedBoards;

  restoreSavedBoard(fanorona,fanorona->latestBoard);
  /*printf("gtk_fanorona_redo(): after: (savedBoards,latestBoard)==(%d,%d)\n\n",
         safeTurnNum(fanorona->savedBoards),
         safeTurnNum(fanorona->latestBoard));*/
}

void
gtk_fanorona_set_click_validator(GtkFanorona* fanorona,
                                 GtkFanoronaClickValidator clickValidator,
                                 void* context)
{
  fanorona->clickValidator.function=clickValidator;
  fanorona->clickValidator.context=context;
}

GtkFanoronaClickValidator
gtk_fanorona_get_click_validator(GtkFanorona* fanorona)
{
  return fanorona->clickValidator.function;
}

void*
gtk_fanorona_get_click_validator_context(GtkFanorona* fanorona)
{
  return fanorona->clickValidator.context;
}

static gboolean moveClickValidator(GtkFanorona* fanorona,
                                   int col,
                                   int row,
                                   void* context)
{
  FanoronaPlayer* player=(FanoronaPlayer*)context;
  FanoronaBoardCellState state=fanoronaBoardGetCell(fanorona->board,
                                                    col,row);

  if (fanorona->ambiguousMoveInProgress)
    {
      GtkFanoronaMove move=fanorona->ambiguousMove;

      int dc,dr,colApproach,rowApproach,colWithdrawal,rowWithdrawal;
      dc=move.colTo-move.colFrom;
      dr=move.rowTo-move.rowFrom;

      colApproach=move.colTo+dc;
      rowApproach=move.rowTo+dr;
      colWithdrawal=move.colFrom-dc;
      rowWithdrawal=move.rowFrom-dr;

      if ((col==colApproach) && (row==rowApproach))
        return TRUE;

      if ((col==colWithdrawal) && (row==rowWithdrawal))
        return TRUE;
    }

  if ((fanorona->moveInProgress.colFrom)<0)
    return (state==(fanorona->curPlayer));

  return fanoronaPlayerOtherPlayerMoveValid(player,
                                            fanorona->moveInProgress.colFrom,
                                            fanorona->moveInProgress.rowFrom,
                                            col,row,
                                            0,0);
}

static void whoseTurn(GtkFanorona* fanorona,
                      gboolean isPlayer1)
{
  gtk_fanorona_set_cur_player(fanorona,
                              (isPlayer1
                               ? fanoronaBoardCellStatePlayer1
                               : fanoronaBoardCellStatePlayer2
                               )
                              );
}

void gtk_fanorona_enable_moves(GtkFanorona* fanorona,
                               FanoronaPlayer* player)
{
  if (!player)
    player=fanoronaPlayerNew(0,3);

  gtk_fanorona_set_board(fanorona,fanoronaPlayerGetBoard(player));

  gtk_signal_connect (GTK_OBJECT (fanorona), "fanorona_clicked",
                      GTK_SIGNAL_FUNC (clicked), NULL);
  gtk_signal_connect(GTK_OBJECT(fanorona),"fanorona_moved",
                     GTK_SIGNAL_FUNC(moved),NULL);
  gtk_signal_connect(GTK_OBJECT(fanorona),"fanorona_report_whose_move",
                     GTK_SIGNAL_FUNC(whoseTurn),NULL);


  gtk_fanorona_set_state_methods(fanorona,
                                 player,
                                 (gtk_fanorona_get_state_method)fanoronaPlayerGetState,
                                 (gtk_fanorona_restore_state_method)fanoronaPlayerRestoreState);

  gtk_fanorona_set_click_validator(fanorona,moveClickValidator,player);
  clearSavedBoards(fanorona);

  gtk_signal_emit(GTK_OBJECT(fanorona),
                  fanorona_signals[FANORONA_REPORT_WHOSE_MOVE],
                  (gboolean)(!fanoronaPlayerIsPlayer1(player)));
}


void gtk_fanorona_new_game(GtkFanorona* fanorona)
{
  void* context=gtk_fanorona_get_click_validator_context(fanorona);
  FanoronaPlayer* player=(FanoronaPlayer*)context;

  fanoronaPlayerNewGame(player);
  clear_all_possibles(fanorona,TRUE);
  clear_all_useds(fanorona,TRUE);
  fanorona->moveInProgress.colFrom=fanorona->moveInProgress.rowFrom=-1;
  clearSavedBoards(fanorona);
  restoreSavedBoard(fanorona,0);
  gtk_widget_set_sensitive(GTK_WIDGET(fanorona),TRUE);

  gtk_signal_emit(GTK_OBJECT(fanorona),
                  fanorona_signals[FANORONA_REPORT_WHOSE_MOVE],
                  (gboolean)(!fanoronaPlayerIsPlayer1(player)));
}

static void clicked(GtkWidget* widget,
                    gint col,
                    gint row,
                    gpointer data)
{
  FanoronaBoardCellState state;
  GtkFanorona* fanorona=GTK_FANORONA(widget);

  void* context=gtk_fanorona_get_click_validator_context(fanorona);
  FanoronaPlayer* player=(FanoronaPlayer*)context;

  /*printf("clicked(%d,%d)\n",col,row);*/

  /*if (fanorona->firstMoveOfTurn)*/ {
    pushCurBoard(fanorona);
    /*fanorona->firstMoveOfTurn=0;*/
  }

  gtk_fanorona_clear_all_possibles(fanorona);

  if (fanorona->ambiguousMoveInProgress)
    {
      GtkFanoronaMove move=fanorona->ambiguousMove;

      int dc,dr,colApproach,rowApproach,colWithdrawal,rowWithdrawal;
      dc=move.colTo-move.colFrom;
      dr=move.rowTo-move.rowFrom;

      colApproach=move.colTo+dc;
      rowApproach=move.rowTo+dr;
      colWithdrawal=move.colFrom-dc;
      rowWithdrawal=move.rowFrom-dr;

      fanorona->ambiguousMoveInProgress=FALSE;

      if ((col==colApproach) && (row==rowApproach))
        move.byApproach=TRUE;
      else
        {
          if ((col==colWithdrawal) && (row==rowWithdrawal))
            move.byApproach=FALSE;
          else
            return;
        }

      gtk_signal_emit(GTK_OBJECT(fanorona),
                      fanorona_signals[FANORONA_MOVED],
                      (gpointer)(&move));
    }
  else
    {
      if ((fanorona->moveInProgress.colFrom)<0)
        {
          int colFrom=col;
          int rowFrom=row;

          int colTo,rowTo;
          int minCol=colFrom-1,maxCol=colFrom+1,
            minRow=rowFrom-1,maxRow=rowFrom+1;

          if (minCol<0)
            minCol=0;
          if (maxCol>=FANORONA_NUM_COLS)
            maxCol=FANORONA_NUM_COLS-1;
          if (minRow<0)
            minRow=0;
          if (maxRow>=FANORONA_NUM_ROWS)
            maxRow=FANORONA_NUM_ROWS-1;

          fanorona->moveInProgress.colFrom=col;
          fanorona->moveInProgress.rowFrom=row;

          for (colTo=minCol; colTo<=maxCol; colTo++)
            for (rowTo=minRow; rowTo<=maxRow; rowTo++)
              {
                gboolean possible
                  =fanoronaPlayerOtherPlayerMoveValid(player,
                                                      colFrom,rowFrom,
                                                      colTo,rowTo,
                                                      0,0);
                gtk_fanorona_set_possible(fanorona,
                                          colTo,rowTo,
                                          possible);
              }
        }
      else
        {
          GtkFanoronaMove move;
          int canApproach,canWithdraw;

          move.colFrom=fanorona->moveInProgress.colFrom;
          move.rowFrom=fanorona->moveInProgress.rowFrom;
          move.colTo=col;
          move.rowTo=row;

          if (!fanoronaPlayerOtherPlayerMoveValid(player,
                                                  fanorona->moveInProgress.colFrom,
                                                  fanorona->moveInProgress.rowFrom,
                                                  col,row,
                                                  &canApproach,
                                                  &canWithdraw)
              )
            {
              fanorona->moveInProgress.colFrom=-1;
              fanorona->moveInProgress.rowFrom=-1;
              return;
            }

          if (canApproach && canWithdraw)
            {
              int dc,dr;
              gtk_signal_emit(GTK_OBJECT(fanorona),
                              fanorona_signals[FANORONA_PROMPT_FOR_DIRECTION]);
              fanorona->ambiguousMove=move;

              dc=move.colTo-move.colFrom;
              dr=move.rowTo-move.rowFrom;

              gtk_fanorona_set_possible(fanorona,
                                        move.colTo+dc,move.rowTo+dr,
                                        TRUE);
              gtk_fanorona_set_possible(fanorona,
                                        move.colFrom-dc,move.rowFrom-dr,
                                        TRUE);
              fanorona->ambiguousMoveInProgress=TRUE;
            }
          else
            {
              move.byApproach=canApproach;

              gtk_signal_emit(GTK_OBJECT(fanorona),
                              fanorona_signals[FANORONA_MOVED],
                              (gpointer)(&move));
            }

          fanorona->moveInProgress.colFrom=-1;
          fanorona->moveInProgress.rowFrom=-1;
        }
    }
}

static void set_possible(GtkFanorona* fanorona,
                         int col,
                         int row,
                         gboolean possible,
                         gboolean invalidate)
{
  gboolean old=fanorona->possibles[col][row];
  fanorona->possibles[col][row]=possible;
  if (invalidate && (possible!=old))
    gtk_fanorona_invalidate_cell(fanorona,col,row);
}

static void clear_all_possibles(GtkFanorona* fanorona,
                                gboolean invalidate)
{
  int col,row;
  for (col=0; col<FANORONA_NUM_COLS; col++)
    for (row=0; row<FANORONA_NUM_ROWS; row++)
      set_possible(fanorona,col,row,FALSE,invalidate);
}

void gtk_fanorona_clear_all_possibles(GtkFanorona* fanorona)
{
  clear_all_possibles(fanorona,TRUE);
}

void gtk_fanorona_set_possible(GtkFanorona* fanorona,
                               int col,
                               int row,
                               gboolean possible)
{
  set_possible(fanorona,col,row,possible,TRUE);
}

gboolean gtk_fanorona_get_possible(GtkFanorona* fanorona,
                                   int col,
                                   int row)
{
  return fanorona->possibles[col][row];
}

static void set_used(GtkFanorona* fanorona,
                     int col,
                     int row,
                     gboolean used,
                     gboolean invalidate)
{
  gboolean old=fanorona->useds[col][row];
  fanorona->useds[col][row]=used;
  if (invalidate && (used!=old))
    gtk_fanorona_invalidate_cell(fanorona,col,row);
}

static void clear_all_useds(GtkFanorona* fanorona,
                            gboolean invalidate)
{
  int col,row;
  for (col=0; col<FANORONA_NUM_COLS; col++)
    for (row=0; row<FANORONA_NUM_ROWS; row++)
      set_used(fanorona,col,row,FALSE,invalidate);
}

void gtk_fanorona_clear_all_useds(GtkFanorona* fanorona)
{
  clear_all_useds(fanorona,TRUE);
}

void gtk_fanorona_set_used(GtkFanorona* fanorona,
                           int col,
                           int row,
                           gboolean used)
{
  set_used(fanorona,col,row,used,TRUE);
}

gboolean gtk_fanorona_get_used(GtkFanorona* fanorona,
                               int col,
                               int row)
{
  return fanorona->useds[col][row];
}

static gint computerMove(gpointer data)
{
  GtkFanorona* fanorona=(GtkFanorona*)data;
  void* context=gtk_fanorona_get_click_validator_context(fanorona);
  FanoronaPlayer* player=(FanoronaPlayer*)context;

  int turnIsOver;

  int colFrom,rowFrom,colTo,rowTo,byApproach;

  if (fanoronaPlayerMakeMove(player,
                             &colFrom,&rowFrom,&colTo,&rowTo,&byApproach,
                             &turnIsOver)
      <0)
    {
      int tmp=errno;
      fprintf(stderr,"fanoronaPlayerMakeMove() failed: %s\n",
              strerror(tmp));
      gtk_exit(2);
    }

  gtk_fanorona_set_used(fanorona,colFrom,rowFrom,TRUE);

  /*printf("Computer player moved from (%d,%d) to (%d,%d), by %s\n",
         colFrom,rowFrom,colTo,rowTo,
         (byApproach ? "approach" : "withdrawal"));*/

  if (turnIsOver)
    {
      /*fanorona->firstMoveOfTurn=1;*/
      if (!fanoronaBoardGameOver(fanoronaPlayerGetBoard(player)))
        gtk_widget_set_sensitive(GTK_WIDGET(fanorona),TRUE);
      gtk_fanorona_clear_all_useds(fanorona);
      gtk_signal_emit(GTK_OBJECT(fanorona),
                      fanorona_signals[FANORONA_REPORT_WHOSE_MOVE],
                      (gboolean)(!fanoronaPlayerIsPlayer1(player)));
    }

  return !turnIsOver;
}

static void moved(GtkFanorona* fanorona,
                  const GtkFanoronaMove* move,
                  gpointer data)
{
  void* context=gtk_fanorona_get_click_validator_context(fanorona);
  FanoronaPlayer* player=(FanoronaPlayer*)context;

  int turnIsOver;

  /*printf("moved() : (%d,%d) to (%d,%d) by %s\n",
         move->colFrom,move->rowFrom,
         move->colTo,move->rowTo,
         ((move->byApproach) ? "approach" : "withdrawal"));*/

  /*fanoronaBoardPrint(stdout,fanoronaPlayerGetBoard(player));*/

  gtk_fanorona_set_used(fanorona,move->colFrom,move->rowFrom,TRUE);

  if (fanoronaPlayerMoveMade(player,
                             move->colFrom,
                             move->rowFrom,
                             move->colTo,
                             move->rowTo,
                             move->byApproach,
                             &turnIsOver)
      <0)
    {
      int tmp=errno;
      fprintf(stderr,"fanoronaPlayerMoveMade() failed: %s\n",
              strerror(tmp));
      gtk_exit(3);
    }

  if (turnIsOver)
    {
      gtk_widget_set_sensitive(GTK_WIDGET(fanorona),FALSE);
      if (!fanoronaBoardGameOver(fanoronaPlayerGetBoard(player)))
        gtk_timeout_add(gtk_fanorona_get_delay(fanorona),
                        computerMove,fanorona);
      gtk_fanorona_clear_all_useds(fanorona);
      gtk_signal_emit(GTK_OBJECT(fanorona),
                      fanorona_signals[FANORONA_REPORT_WHOSE_MOVE],
                      (gboolean)(fanoronaPlayerIsPlayer1(player)));
    }
  else
    gtk_fanorona_set_possible(fanorona,move->colTo,move->rowTo,TRUE);
}

static void invalidate_all(GtkFanorona* fanorona)
{
  GtkWidget* widget=GTK_WIDGET(fanorona);
  GdkEventExpose event;

  event.type=GDK_EXPOSE;
  event.window=widget->window;
  event.send_event=1;
  event.area.x=0;
  event.area.y=0;
  event.area.width=widget->allocation.width;
  event.area.height=widget->allocation.height;
  event.count=0;

  gdk_event_put((GdkEvent*)&event);
}

int gtk_fanorona_get_delay(GtkFanorona* fanorona)
{
  return fanorona->delay;
}

void gtk_fanorona_set_delay(GtkFanorona* fanorona,
                            int delay)

{
  fanorona->delay=delay;
}

void gtk_fanorona_set_state_methods(GtkFanorona* fanorona,
                                    void* context,
                                    gtk_fanorona_get_state_method get,
                                    gtk_fanorona_restore_state_method restore)
{
  fanorona->stateMethods.context=context;
  fanorona->stateMethods.get=get;
  fanorona->stateMethods.restore=restore;
}

static const char* makeConfigPath(const char* configFile)
{
  const char section[]="fanoronaSaved";
  char* path=(char*)malloc(1 /* = */
                           +strlen(configFile)
                           +2 /* =/ */
                           +strlen(section)
                           +2 /* /\0 */
                           );
  if (!path)
    {
      int tmp=errno;
      fprintf(stderr,"GtkFanorona:makeConfigPath(): can't malloc() path: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  sprintf(path,"=%s=/%s/",configFile,section);
  return path;
}

/* For some reason, gnome_config_sync_file() takes char* rather than
 *  const char*.
 */
static char* makeConfigFile(const char* configFile)
{
  char* file=(char*)malloc(1 /* = */
                           +strlen(configFile)
                           +1 /* = */
                           +1 /* \0 */
                           );
  if (!file)
    {
      int tmp=errno;
      fprintf(stderr,"GtkFanorona:makeConfigFile(): can't malloc() file: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  sprintf(file,"=%s=",configFile);
  return file;
}

static unsigned char hexit(int nibble)
{
  if (nibble<0)
    return 0;

  if (nibble<10)
    return '0'+nibble;

  if (nibble<16)
    return 'a'+(nibble-10);

  return 0;
}

static int unhexNibble(char h)
{
  if ((h>='0') && (h<='9'))
    return h-'0';

  if ((h>='a') && (h<='f'))
    return h-'a'+10;

  return 0;
}

static unsigned char unhex(char high,
                           char low)
{
  return (unhexNibble(high)<<4) | unhexNibble(low);
}

static void config_set_binary(const char* name,
                              const void* data,
                              int len)
{
  int i;
  const unsigned char* dataStr=(const unsigned char*)data;
  char* str;

  if (!data)
    {
      gnome_config_set_string(name,"!");
      return;
    }

  if (len<=0)
    {
      gnome_config_set_string(name,"");
      return;
    }

  str=(char*)malloc(len*2+1);
  if (!str)
    return;

  for (i=0; i<len; i++)
    {
      str[i*2]=hexit(dataStr[i]>>4);
      str[i*2+1]=hexit(dataStr[i]&15);
    }

  str[i*2]=0;

  gnome_config_set_string(name,str);
  /*free(str);*/
}

static int config_get_binary(const char* name,
                             void** data,
                             int* len)
{
  const char* str=gnome_config_get_string(name);
  unsigned char* dataStr;
  int i,slen;

  if (!str)
    {
      (*data)=0;
      (*len)=0;
      return 0;
    }

  slen=strlen(str);
  if (slen%2)
    {
      (*data)=0;
      (*len)=0;
      return 0;
    }

  (*len)=slen/2;

  dataStr=(unsigned char*)malloc(*len); /* not +1, because not 0-term */
  for (i=0; i<(*len); i++)
    dataStr[i]=unhex(str[i*2],str[i*2+1]);

  (*data)=dataStr;
  return 0;
}

static int save_ith_savedBoard(const char* curPrefix,
                               GtkFanorona_SavedBoard* board,
                               int i)
{
  char prefix[MAXPATHLEN*2];
  sprintf(prefix,"%ssavedBoard_%d/",curPrefix,i);
  prefix[strlen(curPrefix)-1]='_';

  gnome_config_push_prefix(prefix);

  {
    int col,row;
    for (col=0; col<FANORONA_NUM_COLS; col++)
      for (row=0; row<FANORONA_NUM_ROWS; row++)
        {
          char name[256];
          sprintf(name,"board.cells.%d.%d",
                  col,row);
          gnome_config_set_int(name,(int)(board->board.cells[row][col]));
        }
  }

  gnome_config_set_int("board.numForPlayer1",board->board.numForPlayer1);
  gnome_config_set_int("board.numForPlayer2",board->board.numForPlayer2);
  gnome_config_set_int("board.blockWatchers",board->board.blockWatchers);

  gnome_config_set_int("pressedIn.col",board->pressedIn.col);
  gnome_config_set_int("pressedIn.row",board->pressedIn.row);

  gnome_config_set_int("curPlayer",board->curPlayer);

  gnome_config_set_int("moveInProgress.colFrom",
                       board->moveInProgress.colFrom);
  gnome_config_set_int("moveInProgress.rowFrom",
                       board->moveInProgress.rowFrom);

  {
    int col,row;
    for (col=0; col<FANORONA_NUM_COLS; col++)
      for (row=0; row<FANORONA_NUM_ROWS; row++)
        {
          char name[256];
          sprintf(name,"possibles.%d.%d",col,row);
          gnome_config_set_bool(name,board->possibles[col][row]);

          sprintf(name,"useds.%d.%d",col,row);
          gnome_config_set_bool(name,board->useds[col][row]);
        }
  }

  gnome_config_set_int("ambiguousMove.colFrom",
                       board->ambiguousMove.colFrom);
  gnome_config_set_int("ambiguousMove.rowFrom",
                       board->ambiguousMove.rowFrom);
  gnome_config_set_int("ambiguousMove.colTo",
                       board->ambiguousMove.colTo);
  gnome_config_set_int("ambiguousMove.rowTo",
                       board->ambiguousMove.rowTo);
  gnome_config_set_int("ambiguousMove.byApproach",
                       board->ambiguousMove.byApproach);
  gnome_config_set_bool("ambiguousMoveInProgress",
                        board->ambiguousMoveInProgress);
  /*gnome_config_set_bool("firstMoveOfTurn",
                           board->firstMoveOfTurn);*/

  config_set_binary("externState",
                    board->externState.data,
                    board->externState.len);

  gnome_config_pop_prefix();
  return 0;
}

int gtk_fanorona_save(GtkFanorona* fanorona,
                      const char* configFile)
{
  const char* path=makeConfigPath(configFile);
  char* file;

  if (!path)
    {
      int tmp=errno;
      fprintf(stderr,"gtk_fanorona_save(): can't make config path: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  file=makeConfigFile(configFile);

  if (!file)
    {
      int tmp=errno;
      free((char*)path); /* cast away const */
      fprintf(stderr,"gtk_fanorona_save(): can't make config file: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  gnome_config_push_prefix(path);

  pushCurBoard(fanorona);

  {
    int i;
    GtkFanorona_SavedBoard* cur;
    for (i=1, cur=fanorona->savedBoards; cur; i++, cur=cur->next)
      save_ith_savedBoard(path,cur,i);

    gnome_config_set_int("numSavedBoards",i-1);
  }

  {
    int latestBoardTurnNum;
    if (fanorona->latestBoard)
      latestBoardTurnNum=fanorona->latestBoard->turnNum;
    else
      latestBoardTurnNum=-1;
    gnome_config_set_int("latestBoardTurnNum",latestBoardTurnNum);
  }

  popSavedBoard(fanorona);

  gnome_config_pop_prefix();
  gnome_config_sync_file(file);

  free(file);
  free((char*)path); /* cast away const */

  return 0;
}

static GtkFanorona_SavedBoard* load_ith_savedBoard(const char* curPrefix,
                                                   int i)
{
  GtkFanorona_SavedBoard* board;
  char prefix[MAXPATHLEN*2];

  board=(GtkFanorona_SavedBoard*)malloc(sizeof(GtkFanorona_SavedBoard));
  if (!board)
    {
      int tmp=errno;
      fprintf(stderr,"load_ith_savedBoard(): can't malloc() board: %s\n",
              strerror(tmp));
      errno=tmp;
      return 0;
    }

  sprintf(prefix,"%ssavedBoard_%d/",curPrefix,i);
  prefix[strlen(curPrefix)-1]='_';

  gnome_config_push_prefix(prefix);

  {
    int col,row;
    for (col=0; col<FANORONA_NUM_COLS; col++)
      for (row=0; row<FANORONA_NUM_ROWS; row++)
        {
          char name[256];
          sprintf(name,"board.cells.%d.%d",
                  col,row);
          board->board.cells[row][col]
            =(FanoronaBoardCellState)gnome_config_get_int(name);
        }
  }

  board->board.numForPlayer1=gnome_config_get_int("board.numForPlayer1");
  board->board.numForPlayer2=gnome_config_get_int("board.numForPlayer2");
  board->board.blockWatchers=gnome_config_get_int("board.blockWatchers");

  board->pressedIn.col=gnome_config_get_int("pressedIn.col");
  board->pressedIn.row=gnome_config_get_int("pressedIn.row");

  board->curPlayer=gnome_config_get_int("curPlayer");

  board->moveInProgress.colFrom=gnome_config_get_int("moveInProgress.colFrom");
  board->moveInProgress.rowFrom=gnome_config_get_int("moveInProgress.rowFrom");

  {
    int col,row;
    for (col=0; col<FANORONA_NUM_COLS; col++)
      for (row=0; row<FANORONA_NUM_ROWS; row++)
        {
          char name[256];
          sprintf(name,"possibles.%d.%d",col,row);
          board->possibles[col][row]=gnome_config_get_bool(name);

          sprintf(name,"useds.%d.%d",col,row);
          board->useds[col][row]=gnome_config_get_bool(name);
        }
  }

  board->ambiguousMove.colFrom=gnome_config_get_int("ambiguousMove.colFrom");
  board->ambiguousMove.rowFrom=gnome_config_get_int("ambiguousMove.rowFrom");
  board->ambiguousMove.colTo=gnome_config_get_int("ambiguousMove.colTo");
  board->ambiguousMove.rowTo=gnome_config_get_int("ambiguousMove.rowTo");
  board->ambiguousMove.byApproach=gnome_config_get_int("ambiguousMove.byApproach");
  board->ambiguousMoveInProgress=gnome_config_get_bool("ambiguousMoveInProgress");
  /*board->firstMoveOfTurn=gnome_config_get_bool("firstMoveOfTurn");*/

  config_get_binary("externState",
                    &(board->externState.data),
                    &(board->externState.len));

  gnome_config_pop_prefix();
  return board;
}

int gtk_fanorona_restore(GtkFanorona* fanorona,
                         const char* configFile)
{
  const char* path=makeConfigPath(configFile);

  if (!path)
    {
      int tmp=errno;
      fprintf(stderr,"gtk_fanorona_restore(): can't make config path: %s\n",
              strerror(tmp));
      errno=tmp;
      return -1;
    }

  gnome_config_push_prefix(path);

  gtk_fanorona_new_game(fanorona);

  {
    int i,N=gnome_config_get_int("numSavedBoards");
    int latestBoardTurnNum=gnome_config_get_int("latestBoardTurnNum");
    fanorona->latestBoard=0;
    for (i=N; i>0; i--)
      {
        GtkFanorona_SavedBoard* cur=load_ith_savedBoard(path,i);
        if (!cur)
          {
            int tmp=errno;
            fprintf(stderr,"gtk_fanorona_restore(): can't load saved board: %s\n",
                    strerror(tmp));
            errno=tmp;
            return -1;
          }

        cur->next=fanorona->savedBoards;
        cur->prev=0;
        if (fanorona->savedBoards)
          fanorona->savedBoards->prev=cur;
        fanorona->savedBoards=cur;
        cur->turnNum=i;
        if ((cur->turnNum)==latestBoardTurnNum)
          fanorona->latestBoard=cur;
      }
  }

  popSavedBoard(fanorona);

  gnome_config_pop_prefix();

  free((char*)path); /* cast away const */

  return 0;
}
